''
''	Interface to help make chipmunk simulations run consistently on all machines.
''	The interface prefers threading, but if it isn't supported on the current
''	platform, or is disabled via -lang qb, this also supports a non-threaded
''	method to keep the simulation consistent.
''
''	Programmed by: Alex "Oz" Barry, 2010 - alex.barry@gmail.com
''	If you have any questions, comments, or anything else about this header, feel
''	free to email me and I'll get back to you as soon as I can.
''
''	You can use this code for any of your projects, as long as this comment header
''	remains as it is.  If you modify this in any way, please note the changes.  If
''	you have any improvements, let me know and I'll add you to the credits.
''

#include "../inc/chipmunk/chipmunk.bi"
#inclib "chipmunk"

#ifndef __EASYCHIPMUNK
#define __EASYCHIPMUNK

#if defined( __FB_DOS__ ) OR ( __FB_LANG__ = "qb" )				' If this is compiled for DOS, or with the -lang qb flag, don't allow threading
#	ifdef CAN_THREAD											' Make sure the programmer isn't trying to trick the header (What sort of asshole would do this?)
#		undef CAN_THREAD										' Destroy any CAN_THREAD define, just in case
#	endif
#else															' If this is compiled for any other platform (win32, *nix) or without the -lang qb flag
#	define CAN_THREAD											' Inform the rest of the header that it can/should use threading to stabilize the simulation
#endif

type ecUpdateFunc as sub( byval self as any ptr )
type easyChipmunkPTR as easyChipmunk ptr
type easyChipmunk
public:
	declare constructor()										' Setup a basic space, and tell the simulation that it should be paused.  If threading is supported, the thread is started here
	declare destructor()										' Release all bodies and shapes attached to the space, and free the space.  Stop the thread is threading is supported.
	
	declare function isUsingThread() as ubyte					' Check to see if this instance of easyChipmunk is using threading - this is defined at compile time.
	
	declare sub simulate()										' Dummy function to advance the simulation.  If threading is supported, this is a dummy function
	declare sub _DOSIMULATIONSTEP()								' Real simulation step.  Do not call this directly for any reason!  Use simulate() instead.
	
	declare property physicsTick() as double					' Get the physics tick rate in hertz (tick per second)
	declare property physicsTick( byval newValue as double )	' Set the physics tick rate, and calculate the necessary delta time.  Modifying this during a simulation is not advised!
	declare property deltaTime() as double						' Get the deltatime step used for the simulation
	
	declare sub lock()											' Lock the simulation mutex.  This is a dummy function is threading is not supported.
	declare sub unlock()										' Unlock the simulation mutex.  This is a dummy function is threading is not supported.
	
#ifdef CAN_THREAD												' Block is only included if threading is supported
	declare static sub physicsThread( byval self as any ptr )	' Local subroutine for the thread
	thread			as any ptr									' Reference to thread
	mutex			as any ptr									' Reference to mutex
	killme			as ubyte									' Flag to tell thread if we are finished simulating or not.
#endif															' End thread-supported block

	space			as cpSpace ptr								' Space that is being simulated
	onUpdate		as ecUpdateFunc								' Updated every cpSpaceStep() call - reference to instance is passed
	
	currentTime		as double									' Time that last simulation step occured
	accumulator		as double									' Number of seconds the simulation is behind
	
	userData		as any ptr									' User data - handy for callback

	pause			as ubyte									' Toggle value to tell simulation to pause or not.  Should be wrapped in the lock() / unlock() members.
private:
	_deltaTime		as cpFloat									' Real value of delta time (use for caculation in cpSpaceStep)
	_physicsTick	as double									' Simulation rate in hertz
end type

declare sub __internal_InitEC() constructor						' Internal init function to insure that chipmunk has be initialized

constructor easyChipmunk()										' * Create a new instance of easyChipmunk
	this.onUpdate = NULL										' Don't use an update function by default
	this.userData = NULL										' Quick and easily accessible user data
	this.space = cpSpaceNew()									' Create a new space.  This will usually need to be setup further after an instance is declared (wrap in lock() / unlock() members)
	this.currentTime = Timer									' Set the currentTime - there could be issues if this isn't set to a time relatively soon to the simulation
	this.accumulator = 0.0										' Reset second/deltatime accumulator
	this.physicsTick = 60.0										' Set the physics tickrate to 60 and calculate the deltatime
	this.pause = 1												' By default, the simulation is paused
#ifdef CAN_THREAD												' Block is only included if threading is supported
	this.killme = 0												' Reset thread kill flag
	this.mutex = mutexcreate()									' Create a new mutex
	this.thread = threadcreate( @physicsThread, @this )			' Create the physics thread
#endif															' End thread-supported block
end constructor

destructor easyChipmunk()										' * Clean up this instance of easyChipmunk
#ifdef CAN_THREAD												' Block is only included if threading is supported
	this.lock()													' Lock the thread
		this.killme = 1											' Set thread kill flag
	this.unlock()												' Unlock the thread
	threadwait( this.thread )									' Wait for the thread to finish up
	mutexdestroy( this.mutex )									' Destroy the mutex
#endif															' End thread-supported block
	cpSpaceFreeChildren( this.space )							' Free all associated bodies and shapes in the space
	cpSpaceFree( this.space )									' Free the space
end destructor

function easyChipmunk.isUsingThread() as ubyte					' * Function to see if threading is supported
	' This probably doesn't have to be a function, but I'm awesome enough that
	' I made it one anyway.
#ifdef CAN_THREAD												' Block is only included if threading is supported
	return 1													' Return "true" - threading is being used!
#else															' Block is only included if threading is NOT supported
	return 0													' Return "false" - threading is not available!
#endif															' End thread-supported block
end function

sub easyChipmunk.simulate()										' * Subroutine to advance the simulation
	' This function is a dummy function on platforms that do support threading,
	' however, it should still be used to make sure there are no problems compiling
	' and using this software on platforms that do not support threading.
#ifndef CAN_THREAD												' Block is only included if threading is NOT supported
	this._DOSIMULATIONSTEP										' Manually do a simulation step
#endif															' End thread-supported block
end sub

sub easyChipmunk._DOSIMULATIONSTEP()							' * Subroutine to manually advance the simulation
	dim as double newTime = Timer()								' Grab the current time
	dim as double ftime = newTime - this.currentTime			' Calculate the elapsed time (since last call, in seconds)
	this.accumulator += ftime									' Add the elapesed time to the accumulator
	if not this.pause then										' Make sure the simulation isn't paused
		while this.accumulator >= this.deltaTime				' Loop while there is at least time for one cpSpaceStep with the current delta time
			cpSpaceStep( this.space, this.deltaTime )			' Do one space step
			if this.onUpdate then this.onUpdate( @this )
			this.accumulator -= this.deltaTime					' Take away one timestep (deltatime) from the accumulator
		wend
	end if
	this.currentTime = newTime									' Update the current time
end sub

property easyChipmunk.physicsTick() as double					' * Property to hide the _physicsTick variable
	return this._physicsTick									' just return the value
end property

property easyChipmunk.physicsTick( byval newValue as double )	' * Property to hide the _physicsTick variable so that _deltatime is calculated properly (wrap this with the lock() / unlock() members)
	this._physicsTick = newValue								' Set the new tick rate in hertz
	this._deltaTime = 1.0 / newValue							' Calculate the correct deltatime value
	this.unlock()
end property

property easyChipmunk.deltaTime() as double						' * Property to hide the _deltatime variable so it isn't monkeyed with (no ability to directly write to the variable)
	return this._deltaTime										' just return the value
end property

sub easyChipmunk.lock()											' * Subroutine to lock the thread's mutex
	' This function is a dummy function on platforms that do not support threading,
	' however, it should still be used to make sure there are no problems compiling
	' and using this software on platforms that do support threading.
#ifdef CAN_THREAD												' Block is only included if threading is supported
	mutexlock( this.mutex )										' Lock the mutex
#endif															' End thread-supported block
end sub

sub easyChipmunk.unlock()										' * Subroutine to unlock the thread's mutex
	' This function is a dummy function on platforms that do not support threading,
	' however, it should still be used to make sure there are no problems compiling
	' and using this software on platforms that do support threading.
#ifdef CAN_THREAD												' Block is only included if threading is supported
	mutexunlock( this.mutex )									' Unlock the mutex
#endif															' End thread-supported block
end sub

#ifdef CAN_THREAD												' Block is only included if threading is supported
sub easyChipmunk.physicsThread( byval self as any ptr )			' * Subroutine that is called as a thread (has to be static to keep fbc/createthread happy)
	dim physics as easyChipmunk ptr								' Create a easyChipmunk pointer reference
	physics = cptr( easyChipmunk ptr, self )					' Convert the self parameter back into an easyChipmunk instance
	dim quitThread as ubyte = 0									' Variable to track whether or not the thread has recieved the kill flag
	while quitThread = 0										' Do a loop while no kill flag has been found
		physics->lock()											' Lock the main thread out from tampering with the physics while it is updated
			physics->_DOSIMULATIONSTEP()						' Manually do one simulation step
			if physics->killme = 1 then quitThread = 1			' If the thread kill flag has been set, then inform the loop to quit
		physics->unlock()										' Unlock the physics mutex, so the main thread can tamper with the physics
		sleep 1, 1												' Sleep the absolute smallest amount the OS allows, so the CPU doesn't overwork itself.
	wend
end sub
#endif

sub __internal_InitEC() constructor
	cpInitChipmunk()
end sub

#endif ' __EASYCHIPMUNK
