/' Copyright (c) 2007 Scott Lembcke
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 '/
 
type cpArbiterPTR as cpArbiter ptr
type cpSpacePTR as cpSpace ptr
type cpCollisionHandlerPTR as cpCollisionHandler ptr

'' Determines how fast penetrations resolve themselves.
extern cp_bias_coef alias "cp_bias_coef" as cpFloat
'' Amount of allowed penetration. Used to reduce vibrating contacts.
extern cp_collision_slop alias "cp_collision_slop" as cpFloat

'' Data structure for contact points.
type cpContact
	as cpVect	p, n
	as cpFloat	dist
	as cpVect	r1, r2
	as cpFloat	nMass, tMass, bounce, jnAcc, jtAcc, jBias, bias
	as cpHashValue	hash
end type

'' Contacts are always allocated in groups.
declare function cpContactInit cdecl alias "cpContactInit" ( byval as cpContact ptr, byval as cpVect, byval as cpVect, byval as cpFloat, byval as cpHashValue ) as cpContact ptr

'' Sum the contact impulses. (Can be used after cpSpaceStep() returns)
declare function cpContactsSumImpulses cdecl alias "cpContactsSumImpulses" ( byval as cpContact ptr, byval as integer ) as cpVect
declare function cpContactsSumImpulsesWithFriction cdecl alias "cpContactsSumImpulsesWithFriction" ( byval as cpContact ptr, byval as integer ) as cpVect

enum cpArbiterState
	cpArbiterStateNormal
	cpArbiterStateFirstColl
	cpArbiterStateIgnore
	cpArbiterStateSleep
	cpArbiterStateCached
end enum

'' Data structure for tracking collisions between shapes.
type cpArbiter
	'' Information on the contact points between the objects.
	as integer					numContacts
	as cpContact ptr			contacts
	'' The two shapes and bodies involved in the collision.
	'' These variables are NOT in the order defined by the collision handler.
	'' Using CP_ARBITER_GET_SHAPES and CP_ARBITER_GET_BODIES will save you from
	'' many headaches
	as cpShape ptr				private_a, private_b
	'' Calculated before calling the pre-solve collision handler
	'' Override them with custom values if you want specialized behavior
	as cpFloat					e, u
	'' Used for surface_v calculations, implementation may change
	as cpVect					surface_vr
	as cpTimestamp				stamp
	as cpCollisionHandlerPTR	handler
	'' Are the shapes swapped in relation to the collision handler?
	as cpBool					swappedColl
	as cpArbiterState			state
end type

'' Arbiters are allocated in large buffers by the space and don't require a destroy function
declare function cpArbiterInit cdecl alias "cpArbiterInit" ( byval as cpArbiter ptr, byval as cpShape ptr, byval as cpShape ptr ) as cpArbiter ptr

'' These functions are all intended to be used internally.
'' Inject new contact points into the arbiter while preserving contact history.
declare sub cpArbiterUpdate cdecl alias "cpArbiterUpdate" ( byval as cpArbiter ptr, byval as cpContact ptr, byval as integer, byval as cpCollisionHandlerPTR, byval as cpShape ptr, byval as cpShape ptr )
'' Precalculate values used by the solver.
declare sub cpArbiterPreStep cdecl alias "cpArbiterPreStep" ( byval as cpArbiter ptr, byval as cpFloat )
declare sub cpArbiterApplyCachedImpulse cdecl alias "cpArbiterApplyCachedImpulse" ( byval as cpArbiter ptr )
'' Run an iteration of the solver on the arbiter.
declare sub cpArbiterApplyImpulse cdecl alias "cpArbiterApplyImpulse" ( byval as cpArbiter ptr, byval as cpFloat )

'' Arbiter Helper Functions
declare function cpArbiterTotalImpulse cdecl alias "cpArbiterTotalImpulse" ( byval as cpArbiter ptr ) as cpVect
declare function cpArbiterTotalImpulseWithFriction cdecl alias "cpArbiterTotalImpulseWithFriction" ( byval as cpArbiter ptr ) as cpVect
declare sub cpArbiterIgnore cdecl alias "cpArbiterIgnore" ( byval as cpArbiter ptr )

sub cpArbiterGetShapes( byval arb as const cpArbiter ptr, byval a as cpShape ptr ptr, byval b as cpShape ptr ptr )
	if arb->swappedColl then
		(*a) = arb->private_b
		(*b) = arb->private_a
	else
		(*a) = arb->private_a
		(*b) = arb->private_b
	end if
end sub
#define CP_ARBITER_GET_SHAPES(arb, a, b) dim as cpShape ptr a, b : cpArbiterGetShapes(arb, @a, @b)

sub cpArbiterGetBodies( byval arb as const cpArbiter ptr, byval a as cpBody ptr ptr, byval b as cpBody ptr ptr )
	CP_ARBITER_GET_SHAPES( arb, shape_a, shape_b )
	(*a) = shape_a->body
	(*b) = shape_b->body
end sub
#define CP_ARBITER_GET_BODIES(arb, a, b) dim a as cpBody ptr, b as cpBody ptr: cpArbiterGetBodies( arb, @a, @b )

function cpArbiterIsFirstContact( byval arb as const cpArbiter ptr ) as integer
	return (arb->state = cpArbiterStateFirstColl)
end function

function cpArbiterGetNormal( byval arb as const cpArbiter ptr, byval i as integer ) as cpVect
	dim n as cpVect = arb->contacts[i].n
	if arb->swappedColl then
		return cpvneg(n)
	end if
	return n
end function

function cpArbiterGetPoint( byval arb as const cpArbiter ptr, byval i as integer ) as cpVect
	return arb->contacts[i].p
end function

